Package: shades
Version: 1.4.0
Date: 2019-08-02
Title: Simple Colour Manipulation
Author: Jon Clayden
Maintainer: Jon Clayden <code@clayden.org>
Description: Functions for easily manipulating colours, creating colour scales and calculating colour distances.
Suggests: testthat, covr, ggplot2
Enhances: colorspace
Encoding: UTF-8
License: BSD_3_clause + file LICENCE
URL: https://github.com/jonclayden/shades
BugReports: https://github.com/jonclayden/shades/issues
RoxygenNote: 6.1.1
NeedsCompilation: no
Packaged: 2019-08-02 14:53:11 UTC; jon
Repository: CRAN
Date/Publication: 2019-08-02 16:00:02 UTC
Built: R 4.6.0; ; 2026-04-25 16:48:08 UTC; unix

Package: shades
Version: 1.5.0
Date: 2026-07-01
Title: Simple Colour Manipulation
Authors@R: c(person("Jon", "Clayden", role=c("cre","aut"),
    email="code@clayden.org", comment=c(ORCID="0000-0002-6608-0619")))
Description: Functions for easily manipulating colours, creating colour scales and calculating colour distances.
Suggests: tinytest, covr
Enhances: colorspace, ggplot2, tinyplot
Encoding: UTF-8
License: BSD_3_clause + file LICENCE
URL: https://github.com/jonclayden/shades
BugReports: https://github.com/jonclayden/shades/issues
Config/roxygen2/version: 8.0.0
Config/roxygen2/markdown: TRUE
NeedsCompilation: no
Packaged: 2026-07-01 21:46:15 UTC; jon
Author: Jon Clayden [cre, aut] (ORCID: <https://orcid.org/0000-0002-6608-0619>)
Maintainer: Jon Clayden <code@clayden.org>
Repository: CRAN
Date/Publication: 2026-07-02 21:50:07 UTC

# Generated by roxygen2: do not edit by hand

S3method("!=",shade)
S3method("==",shade)
S3method("[",shade)
S3method("[<-",shade)
S3method(all.equal,shade)
S3method(c,shade)
S3method(coords,default)
S3method(coords,shade)
S3method(print,shade)
S3method(rep,shade)
S3method(rev,shade)
S3method(shade,character)
S3method(shade,color)
S3method(shade,default)
S3method(shade,matrix)
S3method(shade,shade)
S3method(space,default)
S3method(space,shade)
export("%.)%")
export("%_/%")
export(addmix)
export(brightness)
export(chroma)
export(complement)
export(coords)
export(delta)
export(dichromat)
export(distance)
export(gradient)
export(hue)
export(lightness)
export(opacity)
export(recycle)
export(saturation)
export(scalefac)
export(shade)
export(space)
export(submix)
export(swatch)
export(warp)
import(grDevices)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,rect)

[![Build Status](https://travis-ci.org/jonclayden/shades.svg?branch=master)](https://travis-ci.org/jonclayden/shades) [![codecov](https://codecov.io/gh/jonclayden/shades/branch/master/graph/badge.svg)](https://codecov.io/gh/jonclayden/shades)

# Simple colour manipulation in R 😎

The `shades` package allows colours to be manipulated easily in R. Properties such as brightness and saturation can be quickly queried, changed or varied, and perceptually uniform colour gradients can be constructed. It plays nicely with the pipe operator from the [popular `magrittr` package](https://github.com/tidyverse/magrittr), and fits naturally into that paradigm. It can also be used [with `ggplot2` scales](#interoperability-with-ggplot2).

The package is available on [CRAN](https://cran.r-project.org/package=shades). You can also install the current development version from GitHub using [`devtools`](https://github.com/r-lib/devtools):


```r
# install.packages("devtools")
devtools::install_github("jonclayden/shades")
```

Feedback on the package or suggestions are welcome, either by filing an issue or by email.

## Usage

Colours are represented in R using [CSS-style hex strings](https://en.wikipedia.org/wiki/Web_colors), but there is also a dictionary of predefined named colours such as `"red"` and `"blue"`. Either of these may be passed to most graphics functions, but creating variations on a particular colour can be awkward.

The `shades` package defines a simple class, `shade`, which uses exactly this same convention and is entirely compatible with built-in colours, but it also stores information about the coordinates of the colours in a particular [colour space](https://en.wikipedia.org/wiki/Color_space).


```r
library(shades)
red <- shade("red")
print(unclass(red))
## [1] "red"
## attr(,"space")
## [1] "sRGB"
## attr(,"coords")
##      R G B
## [1,] 1 0 0
```

From here, the package switches between colour spaces as required, allowing various kinds of colour manipulation to be performed straightforwardly. For example, let's find the saturation level of a few built-in colours.


```r
saturation(c("papayawhip","lavenderblush","olivedrab"))
## [1] 0.1647100 0.0588200 0.7535287
```

Now let's consider a colour gradient stepping through two different colour spaces, which we might want to use as a palette or colour scale.


```r
swatch(gradient(c("red","blue"), 5))
```

![plot of chunk gradients](tools/figures/gradients-1.svg)

```r
swatch(gradient(c("red","blue"), 5, space="Lab"))
```

![plot of chunk gradients](tools/figures/gradients-2.svg)

Here, we are using the `swatch` function to visualise a set of colours as a series of squares. Notice the more uniform appearance of the gradient when it traverses through the [Lab colour space](https://en.wikipedia.org/wiki/Lab_color_space).

Similarly, we can create a set of new colours by changing the brightness and saturation levels of some base colours, and make the code more readable by using the [`magrittr` pipe operator](https://github.com/tidyverse/magrittr).


```r
library(shades); library(magrittr)
c("red","blue") %>% brightness(0.6) %>% saturation(seq(0,1,0.25)) %>% swatch
```

![plot of chunk saturation](tools/figures/saturation-1.svg)

This operation takes the original two colours, reduces their brightness to 60%, assigns a whole series of saturation levels to the result, and then passes it to `swatch` for visualisation. Notice that the pipeline is combinative (like the base function `outer`), returning each combination of parameters in a multidimensional array. The final shades are arranged in two rows by `swatch`, for convenience.

Note that `NA` can be used as a pass-through value:


```r
"cornflowerblue" %>% saturation(c(NA,seq(0,1,0.25))) %>% swatch
```

![plot of chunk missing](tools/figures/missing-1.svg)

Any of these gradients can be directly passed to a standard graphical function, to be used as a colour scale. However, when choosing a colour scale, it is helpful to bear in mind that some viewers may have a colour vision deficiency (colour blindness), making it harder for them to distinguish certain colours and therefore to see a continuous scale. The `dichromat` function can be used to simulate this.


```r
rev(grDevices::rainbow(9)) %>% dichromat %>% swatch
```

![plot of chunk dichromat](tools/figures/dichromat-1.svg)

```r
gradient("viridis",9) %>% dichromat %>% swatch
```

![plot of chunk dichromat](tools/figures/dichromat-2.svg)

Here we are using the built-in "viridis" colour map, [developed for Python's `matplotlib`](http://bids.github.io/colormap/), which was specifically designed to appear continuous under as many conditions as possible. When shown with simulated red-blindness, the default for `dichromat`, it is clearly much more interpretable than a typical rainbow palette generated by R's built-in graphics functions.

The package also supports colour mixing, either additively (as with light) or subtractively (as with paint). For example, consider additive mixtures of the three primary RGB colours.


```r
c("red", addmix("red","green"), "green", addmix("green","blue"), "blue") %>% swatch
```

![plot of chunk addmix](tools/figures/addmix-1.svg)

Similarly, we can subtractively combine the three secondary colours.


```r
c("cyan", submix("cyan","magenta"), "magenta", submix("magenta","yellow"), "yellow") %>% swatch
```

![plot of chunk submix](tools/figures/submix-1.svg)

A "light mixture" infix operator, `%.)%`, and a "paint mixture" infix operator, `%_/%`, are also available.


```r
("red" %.)% "green") == "yellow"
## [1] TRUE
("cyan" %_/% "magenta") == "blue"
## [1] TRUE
```

Finally, you can calculate perceptual distances to a reference colour, as in


```r
distance(c("red","green","blue"), "red")
## [1]  0.00000 86.52385 53.07649
```

## Interoperability with ggplot2

The `shades` package can be used with the popular [`ggplot2` graphics library](https://github.com/tidyverse/ggplot2) in different ways, with different levels of integration. Firstly, gradients from this package can be used as `ggplot2` colour scales through the manual scale functions; for example,


```r
library(shades); library(ggplot2)
mtcars$cyl<- factor(mtcars$cyl)
ggplot(mtcars, aes(cyl,mpg,fill=cyl)) + geom_boxplot() + scale_fill_manual(values=gradient("viridis",3))
```

![plot of chunk ggplot](tools/figures/ggplot-1.svg)

This does not require the two packages to know anything about each other, and is flexible and powerful, but it doesn't easily allow existing `ggplot2` scales to be modified using the colour manipulation functions from `shades`. As of `shades` version 1.3.0, it is also possible to call the package's colour property functions directly on palette functions and scales, so that (for example), we can darken all colours in an existing scale slightly:


```r
ggplot(mtcars, aes(cyl,mpg,fill=cyl)) + geom_boxplot() + scale_fill_brewer(type="qual")
```

![plot of chunk scales](tools/figures/scales-1.svg)

```r
ggplot(mtcars, aes(cyl,mpg,fill=cyl)) + geom_boxplot() + lightness(scale_fill_brewer(type="qual"), delta(-20))
```

![plot of chunk scales](tools/figures/scales-2.svg)

Notice here that we have chosen to use the `delta()` function, which is available in all colour property functions, to request a *relative* reduction of 20 to the original lightness of each colour in the scale. We could also have given a literal value to fix the lightness of all colours to a certain level.

## Related packages

The `shades` package aims to bring together a range of colour manipulation tools and make them easy to use. However, there are several other packages available that can do similar things, sometimes in slightly different ways. These include

- the `grDevices` package, which is shipped with R and used as the basis for `shades`;
- the venerable [`colorspace` package](https://cran.r-project.org/package=colorspace), which provides formal colour classes and transformations between spaces;
- [`munsell`](https://cran.r-project.org/package=munsell), which interprets colours in Munsell notation and does some colour manipulation;
- [`viridis`](https://cran.r-project.org/package=viridis) and [`RColorBrewer`](https://cran.r-project.org/package=RColorBrewer), which provide the colour scales from `matplotlib` and ColorBrewer;
- [`dichromat`](https://cran.r-project.org/package=dichromat), which provides another implementation of the `dichromat` function (a duplication which I didn't discover until after writing this package's version!); and
- [`colorblindr`](https://github.com/clauswilke/colorblindr), which provides alternative tools for simulating colour blindness in figures.

This package was also partly influenced by [Colors.jl](https://github.com/JuliaGraphics/Colors.jl), a colour manipulation package for Julia.

# Shades 🕶️

The main branch for this is now for shades v2 which is still a work in progress (finding and fixing bugs).

There's a bunch of bug fixing and optimisation to come, but I've overhauled the API now to:
- use lazy evaluation to allow some efficiencies
- shift a lot of the computation over into numpy for efficiency
- improve the api with some nicer functional patterns around the canvas object

#' @import grDevices
#' @importFrom graphics par rect plot text
.onLoad <- function (libname, pkgname)
{
    # Illuminant E
    E <- coords(warp(shade(.equalEnergy,space="XYZ"), "LMS"))
    
    abc <- sapply(.standardObserver, function(xyz) {
        A <- coords(warp(shade(xyz,space="XYZ"), "LMS"))
        a <- E[,2] * A[,3] - E[,3] * A[,2]
        b <- E[,3] * A[,1] - E[,1] * A[,3]
        c <- E[,1] * A[,2] - E[,2] * A[,1]
        c(a, b, c)
    })
    
    # Coefficients used by dichromat()
    .cache$a <- abc[1,]
    .cache$b <- abc[2,]
    .cache$c <- abc[3,]
    
    # Ratios of components of E, also used by dichromat()
    .cache$Er <- c(E[,3]/E[,2], E[,3]/E[,1], E[,2]/E[,1])
    
    # Register tinytest extension if the package is installed
    if (system.file(package="tinytest") != "")
        tinytest::register_tinytest_extension("shades", "expect_equal_shades")
}

# These are approximations to matplotlib's colour scales, with 16 key colours each
.colmaps <- list(magma=c("#000004", "#0B0924", "#20114B", "#3B0F70", "#57157E", "#721F81", "#8C2981", "#A8327D", "#C43C75", "#DE4968", "#F1605D", "#FA7F5E", "#FE9F6D", "#FEBF84", "#FDDEA0", "#FCFDBF"),
                 inferno=c("#000004", "#0C0826", "#240C4F", "#420A68", "#5D126E", "#781C6D", "#932667", "#AE305C", "#C73E4C", "#DD513A", "#ED6925", "#F8850F", "#FCA50A", "#FAC62D", "#F2E661", "#FCFFA4"),
                 plasma=c("#0D0887", "#330597", "#5002A2", "#6A00A8", "#8405A7", "#9C179E", "#B12A90", "#C33D80", "#D35171", "#E16462", "#ED7953", "#F68F44", "#FCA636", "#FEC029", "#F9DC24", "#F0F921"),
                 viridis=c("#440154", "#481A6C", "#472F7D", "#414487", "#39568C", "#31688E", "#2A788E", "#23888E", "#1F988B", "#22A884", "#35B779", "#54C568", "#7AD151", "#A5DB36", "#D2E21B", "#FDE725"),
                 blues=c("#F7FBFF", "#DEEBF7", "#C6DBEF", "#9ECAE1", "#6BAED6", "#4292C6", "#2171B5", "#08519C", "#08306B"),
                 reds=c("#FFF5F0", "#FEE0D2", "#FCBBA1", "#FC9272", "#FB6A4A", "#EF3B2C", "#CB181D", "#A50F15", "#67000D"),
                 rdbu=c("#053061", "#2166AC", "#4393C3", "#92C5DE", "#D1E5F0", "#F7F7F7", "#FDDBC7", "#F4A582", "#D6604D", "#B2182B", "#67001F"),
                 ylorrd=c("#800026", "#BD0026", "#E31A1C", "#FC4E2A", "#FD8D3C", "#FEB24C", "#FED976", "#FFEDA0", "#FFFFCC"))

.colmapspaces <- list(magma="Lab", inferno="Lab", plasma="Lab", viridis="Lab", blues="sRGB", reds="sRGB", rdbu="sRGB", ylorrd="sRGB")

#' Colour gradients
#' 
#' This function returns a set of colours interpolating between the specified
#' key colours, equally separated in the specified space.
#' 
#' The key colours may be specified explicitly, or else a built-in colour map
#' may be used. The maps available are currently those developed for Python's
#' \code{matplotlib} 2.0, namely \code{"magma"}, \code{"inferno"},
#' \code{"plasma"} and \code{"viridis"}, and certain ColorBrewer palettes,
#' namely \code{"Blues"}, \code{"Reds"}, \code{"YlOrRd"} (yellow-orange-red)
#' and \code{"RdBu"} (red-grey-blue, a balanced diverging scale).
#' 
#' @param shades Two or more colours, in any suitable form (see
#'   \code{\link{shade}}), or a named colour map such as \code{"viridis"}.
#' @param steps An integer giving the number of shades required in the palette.
#'   If \code{NULL}, a function will instead be returned that takes this
#'   argument.
#' @param space The colour space to traverse. Defaults to the current space of
#'   \code{shades}, or \code{"Lab"} for the \code{matplotlib} colour maps, or
#'   \code{"sRGB"} otherwise.
#' @return A character vector of class \code{"shade"} containing the gradient
#'   elements in the specified space, or a palette function.
#' 
#' @examples
#' gradient(c("red","blue"), 5)
#' gradient(c("red","blue"), 5, space="Lab")
#' gradient("viridis", 5)
#' @references
#' \url{https://bids.github.io/colormap/} for the \code{matplotlib} colour maps;
#' \url{https://colorbrewer2.org} for the ColorBrewer ones.
#' @author Jon Clayden <code@@clayden.org>
#' @export
gradient <- function (shades, steps = NULL, space = NULL)
{
    if (length(shades) == 1)
    {
        if (is.character(shades) && tolower(shades) %in% names(.colmaps))
            shades <- warp(shade(.colmaps[[tolower(shades)]]), .colmapspaces[[tolower(shades)]])
        else
            stop("A single-element argument should specify a predefined colour map")
    }
    else
        shades <- shade(shades)
    
    if (is.null(space))
        space <- space(shades)
    else if (space != attr(shades,"space"))
        shades <- warp(shades, space)
    
    if (is.null(steps))
        return (function (n) gradient(shades,n,space))
    
    nShades <- length(shades)
    sourceNames <- .names(shades, allowNull=FALSE)
    names <- rep("", steps)
    locs <- seq(1, nShades, length.out=steps)
    finalCols <- matrix(NA, steps, 3L)
    for (i in seq_along(locs))
    {
        loc <- locs[i]
        fraction <- loc - floor(loc)
        if (fraction < 1e-5)
            names[i] <- sourceNames[floor(loc)]
        else if ((1 - fraction) < 1e-5)
            names[i] <- sourceNames[ceiling(loc)]
        finalCols[i,] <- (1 - fraction) * coords(shades)[floor(loc),] + fraction * coords(shades)[ceiling(loc),]
    }
    
    if (any(nchar(names) > 0))
        rownames(finalCols) <- names
    return (shade(finalCols, space=space))
}

using("shades")

# Create colour scales traversing different spaces
expect_equal_shades(gradient(c("red","blue"),3), c("#FF0000","#800080","#0000FF"))
expect_equal_shades(rev(gradient(c("red","blue"),3)), c("#0000FF","#800080","#FF0000"))
expect_equal_shades(gradient(c("red","blue"),3,space="LAB"), c("#FF0000","#C90089","#0000FF"))
expect_equal_shades(gradient("magma",3), c("#000004","#B63779","#FCFDBF"))
expect_error(gradient("nothing",3), "should specify a predefined colour map")

.mix <- function (base, mixer, op, amount = 1, space = NULL)
{
    op <- match.fun(op)
    
    if (is.null(space))
        space <- space(base)
    
    amount <- rep(amount, length.out=length(mixer))
    
    base <- warp(base, space)
    mixer <- warp(mixer, space)
    
    baseIndices <- rep(seq_along(base), each=length(mixer))
    mixerIndices <- rep(seq_along(mixer), length(base))
    
    coords <- sapply(1:3, function(i) op(coords(base)[baseIndices,i], amount[mixerIndices]*coords(mixer)[mixerIndices,i]))
    if (!is.matrix(coords))
        coords <- matrix(coords, nrow=1)
    
    shade(.clip(coords,space), space=space)
}

#' Complementary colours
#' 
#' This function returns the complement of its argument, the "opposite" colours
#' in the specified space.
#' 
#' @param shades One or more colours, in any suitable form (see
#'   \code{\link{shade}}), or a palette function or scale.
#' @param space The space in which to take the complement.
#' @return New colours of class \code{"shade"}, or a new palette function.
#' 
#' @examples
#' complement("cyan")
#' complement("cyan", space="HSV")
#' complement("cyan", space="Lab")
#' @export
complement <- function (shades, space = NULL)
{
    if (is.function(shades))
        function (...) complement(shades(...), space)
    else if (inherits(shades, "Scale"))
    {
        ggplot2::ggproto(NULL, shades, palette=function(self,...) {
            colours <- ggplot2::ggproto_parent(shades, self)$palette(...)
            complement(colours, space)
        })
    }
    else
    {
        if (is.null(space))
            space <- space(shades)
    
        if (tolower(space) == "hsv")
            hue(shades, delta(180))
        else
        {
            white <- warp("white", space=space)
            .mix(white, shades, "-", space=space)
        }
    }
}

#' Colour mixtures
#' 
#' These functions allow colours to be mixed in any colour space, either
#' additively (like light) or subtractively (like paint). The infix form
#' \code{\%.)\%} is an alternative for \code{addmix}, and \code{\%_/\%} for
#' \code{submix}, with the mixing amount being fixed to 1 in these cases.
#' 
#' @param base,X A vector of base colours, or a palette function or scale.
#' @param mixer,Y A vector of colours to mix in.
#' @param amount The amount of each colour to mix in, relative to the amount
#'   of the base. This will be recycled to the length of \code{mixer}.
#' @param space A string giving the space in which to perform the mixing, or
#'   \code{NULL}. In the latter case, the space of \code{base} will be used.
#' @return New colours of class \code{"shade"}, or a new palette function.
#' 
#' @examples
#' addmix(c("red","green","blue"), "red")
#' submix(c("cyan","magenta","yellow"), "cyan")
#' @author Jon Clayden <code@@clayden.org>
#' @rdname mixtures
#' @export
addmix <- function (base, mixer, amount = 1, space = NULL)
{
    if (is.function(base))
        function (...) addmix(base(...), mixer, amount, space)
    else if (inherits(base, "Scale"))
    {
        ggplot2::ggproto(NULL, base, palette=function(self,...) {
            colours <- ggplot2::ggproto_parent(base, self)$palette(...)
            addmix(colours, mixer, amount, space)
        })
    }
    else
        .mix(base, mixer, "+", amount, space)
}

#' @rdname mixtures
#' @export
submix <- function (base, mixer, amount = 1, space = NULL)
{
    if (is.function(base))
        function (...) submix(base(...), mixer, amount, space)
    else if (inherits(base, "Scale"))
    {
        ggplot2::ggproto(NULL, base, palette=function(self,...) {
            colours <- ggplot2::ggproto_parent(base, self)$palette(...)
            submix(colours, mixer, amount, space)
        })
    }
    else
        complement(.mix(complement(base,space), complement(mixer,space), "+", amount, space))
}

#' @rdname mixtures
#' @export
"%.)%" <- function (X, Y)
{
    addmix(X, Y)
}

#' @rdname mixtures
#' @export
"%_/%" <- function (X, Y)
{
    submix(X, Y)
}

viridis <- gradient("viridis")
expect_equal(viridis(5L), gradient("viridis",5L))

# The distinction is in the parentheses here: in one case lightness() is called on colours generated by the colour ramp; in the other, the ramp function is wrapped by lightness() and then called
expect_equal(lightness(viridis(5L)), lightness(viridis)(5L))
expect_equal(lightness(viridis(5L),50), lightness(viridis,50)(5L))

expect_equal(opacity(viridis(5L),0.5), opacity(viridis,0.5)(5L))
expect_equal(complement(viridis(5L)), complement(viridis)(5L))
expect_equal(addmix(viridis(5L),"red"), addmix(viridis,"red")(5L))
expect_equal(submix(viridis(5L),"red"), submix(viridis,"red")(5L))

if (requireNamespace("ggplot2", quietly=TRUE))
{
    library(ggplot2)
    data <- data.frame(sex=c("M","M","F","F"), age=c(23,34,28,26), height=c(180,168,159,170))
    
    # This relies on scales::viridis_pal (at time of writing) agreeing with our viridis scale, but this seems less fragile than assuming that scales and ggplot2 continue to interact the way they do now (and adding another suggested dependency)
    plot <- ggplot(data, aes(x=age,y=height,colour=sex)) + geom_point() + lightness(scale_colour_viridis_d(), 50)
    expect_true(all(layer_data(plot)$colour %in% lightness(viridis(2), 50)))
    
    plot <- ggplot(data, aes(x=age,y=height,colour=sex)) + geom_point() + opacity(scale_colour_viridis_d(), 0.5)
    expect_true(all(layer_data(plot)$colour %in% opacity(viridis(2), 0.5)))
    
    plot <- ggplot(data, aes(x=age,y=height,colour=sex)) + geom_point() + complement(scale_colour_viridis_d())
    expect_true(all(layer_data(plot)$colour %in% complement(viridis(2),space="sRGB")))
    
    plot <- ggplot(data, aes(x=age,y=height,colour=sex)) + geom_point() + addmix(scale_colour_viridis_d(), "red")
    expect_true(all(layer_data(plot)$colour %in% addmix(viridis(2), "red", space="sRGB")))
    
    plot <- ggplot(data, aes(x=age,y=height,colour=sex)) + geom_point() + submix(scale_colour_viridis_d(), "red")
    expect_true(all(layer_data(plot)$colour %in% submix(viridis(2), "red", space="sRGB")))
}

using("shades")

# Colour mixing and complements
expect_equivalent(coords(complement("cyan")), coords(shade("red")))
expect_equal(complement("cyan",space="HSV"), warp("red","HSV"))
expect_equal_shades(complement("cyan",space="Lab"), "#4F0002")

expect_equivalent(addmix("blue","green"), shade("#00FFFF"))
expect_equivalent(submix("cyan","yellow"), shade("#00FF00"))
expect_equivalent("blue" %.)% "green", shade("#00FFFF"))
expect_equivalent("cyan" %_/% "yellow", shade("#00FF00"))

# Linearised transformation matrices
.bradfordXYZtoLMS <- matrix(c(0.8951, -0.7502, 0.0389, 0.2664, 1.7135, -0.0685, -0.1614, 0.0367, 1.0296), 3, 3)
.bradfordLMStoXYZ <- solve(.bradfordXYZtoLMS)

# Oklab reference https://bottosson.github.io/posts/oklab/
.XYZtoOklabM1 <- matrix(c(0.8189330101, 0.0329845436, 0.0482003018, 0.3618667424, 0.9293118715, 0.2643662691, -0.1288597137, 0.0361456387, 0.6338517070), 3, 3)
.XYZtoOklabM2 <- matrix(c(0.2104542553, 1.9779984951, 0.0259040371, 0.7936177850, -2.4285922050, 0.7827717662, -0.0040720468, 0.4505937099, -0.8086757660), 3, 3)
.OklabtoXYZM1 <- solve(.XYZtoOklabM1)
.OklabtoXYZM2 <- solve(.XYZtoOklabM2)

# Standard and additional colour space converters
.converters <- list(rgb="sRGB", srgb="sRGB", xyz="XYZ", "apple rgb"="Apple RGB", "cie rgb"="CIE RGB", lab="Lab", luv="Luv")

# Since R 3.6.0, colorConverter() has a "vectorized" argument that allows converters to indicate that they can handle multiple colours at once. For now we don't use that, for backwards compatibility
.converters$hsv <- colorConverter(
    toXYZ = function (hsv, ...) {
        c <- hsv[2] * hsv[3]            # chroma
        base <- (1 - hsv[2]) * hsv[3]
        sextile <- (hsv[1] %% 360) / 60 + 1
        f <- sextile - floor(sextile)
        rgb <- base + switch(as.integer(sextile), c(c,f*c,0), c((1-f)*c,c,0), c(0,c,f*c), c(0,(1-f)*c,c), c(f*c,0,c), c(c,0,(1-f)*c))
        colorspaces$sRGB$toXYZ(rgb, ...)
    },
    fromXYZ = function (xyz, ...) {
        # This rounding operation mirrors convertColor(), and avoids numerical variability between platforms
        rgb <- round(colorspaces$sRGB$fromXYZ(xyz, ...), 5)
        rgb[rgb < 0] <- 0
        rgb[rgb > 1] <- 1
        rgb <- (if (is.matrix(rgb)) t(rgb) else matrix(rgb,ncol=1L))
        hsv <- drop(rgb2hsv(rgb, maxColorValue=1))
        hsv[1] <- (hsv[1] * 360) %% 360
        structure(hsv, names=c("H","S","V"))
    },
    name = "HSV")

.converters$lms <- colorConverter(
    toXYZ = function (lms, ...) { .bradfordLMStoXYZ %*% lms },
    fromXYZ = function (xyz, ...) { structure(.bradfordXYZtoLMS %*% xyz, names=c("L","M","S")) },
    name = "LMS")

.converters$lch <- colorConverter(
    toXYZ = function (lch, ...) {
        angle <- lch[3] / 180 * pi
        lab <- c(lch[1], lch[2] * cos(angle), lch[2] * sin(angle))
        colorspaces$Lab$toXYZ(lab, ...)
    },
    fromXYZ = function (xyz, ...) {
        lab <- colorspaces$Lab$fromXYZ(xyz, ...)
        lch <- c(lab[1], sqrt(lab[2]^2 + lab[3]^2), atan2(lab[3],lab[2]) / pi * 180)
        lch[3] <- lch[3] %% 360
        structure(lch, names=c("L","C","h"))
    },
    name = "LCh")

.converters$oklab <- colorConverter(
    toXYZ = function (oklab, ...) {
        .OklabtoXYZM1 %*% (.OklabtoXYZM2 %*% oklab)^3
    },
    fromXYZ = function (xyz, ...) {
        # x^(1/3) returns NaN for negative numbers, so a bit more care is needed
        lab <- .XYZtoOklabM1 %*% xyz
        labPrime <- sign(lab) * abs(lab)^(1/3)
        structure(.XYZtoOklabM2 %*% labPrime, names=c("L","a","b"))
    },
    white = "D65",
    name = "Oklab")

.toHex <- function (coords, space, alpha = NULL)
{
    space <- tolower(space)
    missing <- apply(coords, 1, anyNA)
    
    if (any(!missing) && !identical(.converters[[space]], "sRGB"))
        coords[!missing,] <- convertColor(coords[!missing,,drop=FALSE], .converters[[space]], "sRGB")
    
    result <- rep(NA_character_, nrow(coords))
    if (is.null(alpha) || all(alpha == 1))
        result[!missing] <- rgb(coords[!missing,1], coords[!missing,2], coords[!missing,3], maxColorValue=1)
    else
        result[!missing] <- rgb(coords[!missing,1], coords[!missing,2], coords[!missing,3], pmax(0,pmin(1,alpha[!missing])), maxColorValue=1)
    return (result)
}

.clip <- function (coords, space)
{
    if (grepl("rgb$", tolower(space)))
    {
        coords[coords < 0] <- 0
        coords[coords > 1] <- 1
    }
    else if (tolower(space) == "hsv")
    {
        temp <- coords[,1] %% 360
        coords[coords < 0] <- 0
        coords[coords > 1] <- 1
        coords[,1] <- temp
    }
    
    return (coords)
}

.dims <- function (x, collapse = FALSE)
{
    if (is.null(dim(x)))
        return (length(x))
    else if (length(dim(x)) > 2 && collapse)
        return (c(dim(x)[1], prod(dim(x)[-1])))
    else
        return (dim(x))
}

.hasNames <- function (x) !is.null(names(x)) && any(nchar(names(x)) > 0)

.names <- function (x, ..., allowNull = TRUE)
{
    if (!missing(..1))
    {
        elements <- lapply(list(x,...), .names, allowNull=FALSE)
        result <- do.call("c", elements)
    }
    else if (!is.null(names(x)))
        result <- names(x)
    else
        result <- rep("", length(x))

    if (allowNull && (is.null(result) || all(result == "")))
        return (NULL)
    else
        return (result)
}

.alpha <- function (x, ..., allowNull = TRUE)
{
    if (!missing(..1))
    {
        elements <- lapply(list(x,...), .alpha, allowNull=FALSE)
        result <- do.call("c", elements)
    }
    else
    {
        if (!is.null(attr(x, "alpha")))
            result <- pmax(0, pmin(1, attr(x,"alpha")))
        else if (any(grepl("#[0-9A-Fa-f]{8}", as.character(x), perl=TRUE)))
            result <- unname(col2rgb(as.character(x), alpha=TRUE)["alpha",] / 255)
        else
            result <- rep(1, length(x))
    }
    
    if (allowNull && all(result == 1))
        return (NULL)
    else
        return (result)
}

#' The shade class
#' 
#' Objects of class `"shade"` are simply standard R character vectors
#' representing one or more 8-bit (s)RGB colours in CSS-like hex format, but
#' with extra attributes giving the current colour space and coordinates.
#' Opacity values are also supported.
#'
#' The `shade()` function creates a colour vector from a single main argument
#' plus method-specific arguments where appropriate. The `shades()` function
#' provides an alternative interface that converts each of its arguments to
#' shades, with direct inline naming, and then concatenates them.
#' 
#' Comparison between `"shade"` objects `x` and `y` is achieved by converting
#' `y` (the second argument) into the colour space of `x` and then comparing
#' coordinates, after any clipping.
#' 
#' @param x,y R objects, or `"shade"` objects for methods.
#' @param space For a matrix, the space in which coordinates are being
#'   provided.
#' @param alpha For a matrix, an associated vector of opacity values between 0
#'   and 1, if required.
#' @param target,current Shade vectors to compare.
#' @param i An index vector.
#' @param value A vector of replacement colours.
#' @param hexonly If `TRUE`, compare only on the basis of the hex strings.
#'   Otherwise test for equal coordinates.
#' @param ... For `shade()`, additional parameters to methods. For `shades()`
#'   and `c()`, any number of colours in any acceptable form.
#' @return A character vector of class `"shade"`, with additional attributes
#'   as follows.
#'     \item{space}{A string naming a colour space.}
#'     \item{coords}{A matrix giving colour coordinates in the relevant space,
#'       one colour per row.}
#'     \item{alpha}{A numeric vector of alpha (opacity) values between 0 and 1,
#'       one value per element of the vector. If not present then full opacity
#'       is assumed for every element.}
#'     \item{names}{A character vector of element names, in the usual R
#'       convention. If not present then the object is unnamed.}
#' 
#' @note When concatenating, shades that are all from the same space will
#'   remain in that space, but shades from different spaces will be warped to
#'   "XYZ" space.
#' 
#' @examples
#' s <- shade(c("red", "green", "blue"))
#' s[1]
#' s[1] <- "pink"
#' @author Jon Clayden <code@@clayden.org>
#' @export
shade <- function (x, ...)
{
    # Zero-length special case - nothing to dispatch on
    if (nargs() == 0)
        return (structure(character(0), space="sRGB", coords=matrix(NA,0,3), class="shade"))
    
    UseMethod("shade")
}

#' @rdname shade
#' @export
shade.shade <- function (x, ...)
{
    return (x)
}

#' @rdname shade
#' @export
shade.color <- function (x, ...)
{
    hex <- colorspace::hex(x, fixup=TRUE)
    coords <- colorspace::coords(x)
    structure(hex, names=rownames(coords), space=class(x), coords=coords, alpha=.alpha(hex), class="shade")
}

#' @rdname shade
#' @export
shade.matrix <- function (x, space = "sRGB", alpha = NULL, ...)
{
    hex <- .toHex(x, space, alpha)
    structure(hex, names=rownames(x), space=space, coords=x, alpha=.alpha(hex), class="shade")
}

#' @rdname shade
#' @export
shade.character <- function (x, ...)
{
    if (length(x) == 0)
        return (shade())
    
    coords <- structure(t(col2rgb(x)/255), dimnames=list(NULL,c("R","G","B")))
    coords[is.na(x),] <- NA
    alpha <- .alpha(x)
    
    nonHex <- !is.na(x) & !grepl("^\\s*#", x, perl=TRUE)
    if (!.hasNames(x) && any(nonHex))
        names(x) <- ifelse(nonHex, x, "")
    
    structure(.toHex(coords,"sRGB",alpha), names=names(x), space="sRGB", coords=coords, alpha=alpha, class="shade")
}

#' @rdname shade
#' @export
shade.default <- function (x, ...)
{
    shade.character(as.character(x), ...)
}

#' @rdname shade
#' @export
shades <- function (...)
{
    args <- list(...)
    elements <- lapply(unname(args), shade)
    result <- do.call(c, elements)
    
    if (.hasNames(args))
    {
        resultNames <- .names(result, allowNull=FALSE)
        lengths <- vapply(elements, length, integer(1L))
        argNames <- rep(.names(args, allowNull=FALSE), lengths)
        indices <- which(argNames != "")
        names(result) <- replace(resultNames, indices, argNames[indices])
    }
    
    return (result)
}

#' @rdname shade
#' @export
print.shade <- function (x, ...)
{
    len <- length(x)
    hasAlpha <- !is.null(attr(x, "alpha"))
    cat(paste0(" ", len, ifelse(.hasNames(x)," named",""), ifelse(len==1," shade"," shades"), " in ", space(x), " space, ", ifelse(hasAlpha,"with","without"), " transparency\n"))
    print(structure(x, space=NULL, coords=NULL, alpha=NULL, class=NULL), quote=FALSE)
}

#' @rdname shade
#' @export
"[.shade" <- function (x, i)
{
    structure(as.character(x)[i], space=attr(x,"space"), coords=attr(x,"coords")[i,,drop=FALSE], alpha=attr(x,"alpha")[i], class="shade")
}

#' @rdname shade
#' @export
"[<-.shade" <- function (x, i, value)
{
    space <- attr(x, "space")
    replacement <- warp(value, space)
    coords <- attr(x, "coords")
    coords[i,] <- attr(replacement, "coords")
    
    alpha <- .alpha(x, allowNull=FALSE)
    alpha[i] <- .alpha(replacement, allowNull=FALSE)
    if (all(alpha == 1))
        alpha <- NULL
    
    names <- .names(x, allowNull=FALSE)
    names[i] <- .names(replacement, allowNull=FALSE)
    if (all(names == ""))
        names <- NULL
    
    structure(.toHex(coords,space,alpha), space=space, coords=coords, alpha=alpha, names=names, class="shade")
}

#' @rdname shade
#' @export
c.shade <- function (...)
{
    shades <- lapply(list(...), shade)
    spaces <- sapply(shades, space)
    
    if (all(spaces == spaces[1]))
        space <- spaces[1]
    else
    {
        space <- "XYZ"
        shades <- lapply(shades, warp, "XYZ")
    }
    
    structure(do.call("c",lapply(shades,as.character)), space=space, coords=do.call("rbind",lapply(shades,coords)), alpha=do.call(".alpha",shades), names=do.call(".names",shades), class="shade")
}

#' @rdname shade
#' @export
rep.shade <- function (x, ...)
{
    indices <- rep(seq_along(x), ...)
    structure(as.character(x)[indices], space=attr(x,"space"), coords=attr(x,"coords")[indices,,drop=FALSE], alpha=attr(x,"alpha")[indices], names=.names(x)[indices], class="shade")
}

#' @rdname shade
#' @export
rev.shade <- function (x)
{
    indices <- rev(seq_along(x))
    structure(as.character(x)[indices], space=attr(x,"space"), coords=attr(x,"coords")[indices,,drop=FALSE], alpha=attr(x,"alpha")[indices], names=.names(x)[indices], class="shade")
}

#' @rdname shade
#' @export
"==.shade" <- function (x, y)
{
    y <- rep(warp(y,attr(x,"space")), length.out=length(x))
    xCoords <- coords(x)
    yCoords <- coords(y)
    coordsAgree <- sapply(seq_along(x), function(i) all(xCoords[i,] == yCoords[i,]))
    alphaAgrees <- .alpha(x, allowNull=FALSE) == .alpha(y, allowNull=FALSE)
    return (coordsAgree & alphaAgrees)
}

#' @rdname shade
#' @export
"!=.shade" <- function (x, y)
{
    return (!`==.shade`(x,y))
}

#' @rdname shade
#' @method all.equal shade
#' @exportS3Method all.equal shade
all.equal.shade <- function (target, current, hexonly = FALSE, ...)
{
    if (hexonly)
        all.equal(as.character(target), as.character(current), ...)
    else if (length(target) != length(current))
        paste0("Lengths do not match (", length(target), " and ", length(current), ")")
    else if (!all(.alpha(target,allowNull=FALSE) == .alpha(current,allowNull=FALSE)))
        paste0("Alpha values do not match (mean absolute difference is ", signif(mean(abs(.alpha(target,allowNull=FALSE) - .alpha(current,allowNull=FALSE))),4), ")")
    else
    {
        target <- warp(target, space(current))
        result <- all.equal(coords(target), coords(current), ...)
        if (isTRUE(result))
            return (TRUE)
        else
        {
            distances <- sapply(seq_along(target), function(i) distance(target[i],current[i]))
            paste0("Mean colour distance is ", signif(mean(distances,4)))
        }
    }
}

#' Retrieve the space of a colour vector
#' 
#' This function retrieves the colour space in which its argument is currently
#' defined.
#' 
#' @param x An R object.
#' @param ... Additional arguments to methods.
#' @return A string naming a colour space.
#'
#' @examples
#' space("red")
#' @author Jon Clayden <code@@clayden.org>
#' @export
space <- function (x, ...)
{
    UseMethod("space")
}

#' @export
space.shade <- function (x, ...)
{
    attr(x, "space")
}

#' @export
space.default <- function (x, ...)
{
    space.shade(shade(x, ...))
}

#' Retrieve the coordinates of a colour vector
#' 
#' This function retrieves the coordinates of a colour vector's elements,
#' within whatever space it is currently defined.
#' 
#' @param x An R object.
#' @param ... Additional arguments to methods.
#' @return A matrix giving colour coordinates in the relevant space, one colour
#'   per row. Columns are typically named.
#'
#' @examples
#' coords("red")
#' @author Jon Clayden <code@@clayden.org>
#' @export
coords <- function (x, ...)
{
    UseMethod("coords")
}

#' @export
coords.shade <- function (x, ...)
{
    attr(x, "coords")
}

#' @export
coords.default <- function (x, ...)
{
    coords.shade(shade(x, ...))
}

#' Shift colours between spaces
#' 
#' This function shifts the current colour space of its arguments to the
#' specified space, returning a new object of class \code{"shade"}.
#' 
#' Valid names for spaces are currently those supported by the
#' \code{\link{convertColor}} function, namely "sRGB", "Apple RGB", "CIE
#' RGB", "XYZ", "Lab" and "Luv"; plus "RGB" (which is treated as an
#' alias for "sRGB"), "HSV", "LCh", "LMS" and "Oklab". Case is not significant.
#' 
#' @param x An R object which can be coerced to class \code{"shade"}.
#' @param space A string naming the new space.
#' @return A new object of class \code{"shade"}.
#' 
#' @note LMS space, used for chromatic adaptation and simulating colour
#'   blindness, is not uniquely defined. Here we use the (linearised) Bradford
#'   transform, obtained by Lam (1985) and used widely in ICC colour profiles
#'   and elsewhere, to transform to and from CIE XYZ space.
#'   
#'   R uses the D65 standard illuminant as the reference white for the "Lab"
#'   and "Luv" spaces.
#' 
#' @examples
#' warp("red", "HSV")
#' @references
#' The Oklab colour space is explained at \url{https://bottosson.github.io/posts/oklab/}.
#' 
#' Lam, K.M. (1985). Metamerism and colour constancy. PhD thesis, University of
#' Bradford.
#' @seealso \code{\link{convertColor}}
#' @author Jon Clayden <code@@clayden.org>
#' @export
warp <- function (x, space)
{
    x <- shade(x)
    sourceSpace <- tolower(attr(x, "space"))
    targetSpace <- tolower(space)
    
    if (sourceSpace == targetSpace)
        return (x)
    
    coords <- attr(x, "coords")
    missing <- apply(coords, 1, anyNA)
    if (any(!missing))
        coords[!missing,] <- convertColor(coords[!missing,,drop=FALSE], .converters[[sourceSpace]], .converters[[targetSpace]])
    alpha <- .alpha(x)
    
    return (structure(.toHex(coords,targetSpace,alpha), names=names(x), dim=dim(x), space=space, coords=coords, alpha=alpha, class="shade"))
}

using("shades")

# Missing shade handling
shades <- gradient(c("red","blue"), 3)
shades[2] <- NA

# Note implicit naming
expect_equal(is.na(shades), c(red=FALSE,TRUE,blue=FALSE))
expect_true(all(is.na(coords(shades)[2,])))
expect_equal(saturation(shades), c(1,NA,1))
expect_equal_shades(saturation(shades,0.5), shade(c("#FF8080",NA,"#8080FF")))
expect_equivalent(is.na(saturation(shades, c(0,0.5,1))), matrix(c(FALSE,TRUE,FALSE),3,3,byrow=TRUE))
expect_equivalent(complement(shades), shade(c("#00FFFF",NA,"#FFFF00")))
expect_equivalent(dichromat(shades)[2], shade(NA))
expect_equal_shades(opacity(shades,0.5), c("#FF000080",NA,"#0000FF80"))

# NAs as new property values should lead to pass-through
expect_equal(saturation(saturation("olivedrab",NA)), saturation("olivedrab"))
expect_equal_shades(opacity("red",NA), "red")
expect_equal_shades(opacity("red",c(0,NA,1)), c("#FF000000","#FF0000FF","#FF0000FF"))

# Create shade objects from various objects and none
expect_length(shade(), 0L)
expect_equal(shade(), shade(character(0)))
expect_equal(names(shade("red")), "red")
expect_stdout(print(shade("red")), "1 named shade")
expect_equal(space(shade("red")), "sRGB")
expect_equivalent(coords("red"), matrix(c(1,0,0),nrow=1))
expect_equivalent(shade(matrix(c(1,0,0),nrow=1)), shade("#FF0000"))
# Using a factor to check the default initialisation method
expect_equivalent(shade(factor(rep("red",3))), rep(shade("red"),3))

shade <- shade("red")
shade[1] <- "#0bead5"
expect_null(names(shade))

# Try the shades() interface
randomRgbCoords <- matrix(runif(6), nrow=2)
expect_equal(shades(red="#ee0000", random=randomRgbCoords),
             c(shade(c(red="#ee0000")), structure(shade(randomRgbCoords), names=rep("random",2))))

if (requireNamespace("colorspace", quietly=TRUE))
    expect_true(shade(colorspace::sRGB(1,0,0)) == "red")

# Convert shade objects between spaces
expect_equivalent(coords(warp("red","HSV")), matrix(c(0,1,1),nrow=1))
expect_equivalent(round(coords(warp(shade("red"),"LAB"))), matrix(c(53,80,67),nrow=1))
# Check that precision loss isn't too great when making a round-trip conversion
shade <- shade(matrix(runif(3),nrow=1), space="sRGB")
expect_equivalent(coords(warp(warp(shade,"HSV"),"sRGB")), coords(shade), tolerance=1e-4)

# Index, combine and compare shade objects
shades <- shade(c("red","green","blue"))
expect_true(shades[1] == "red")
shades[1] <- "darkred"
expect_equal(names(shades)[1], "darkred")
expect_false(shades[1] == "red")

expect_equal(c(shade("red"),shade("green")), shade(c("red","green")))
expect_true(shade("red") != shade("green"))
expect_true(shade("red") == shade(matrix(c(0,1,1),nrow=1),space="HSV"))
expect_equal(space(c(shade("red"), shade(matrix(c(0,1,1),nrow=1),space="HSV"))), "XYZ")

expect_match(all.equal(shade("red"),shade("green")), "Mean colour distance is")
expect_match(all.equal(shade("red"),shade(c("green","blue"))), "Lengths do not match")

.makeDimnames <- function (source, result, labels, firstName)
{
    if (is.null(dim(result)) || length(dim(result)) <= max(1,length(dim(source))) || is.null(labels))
        return (NULL)
    
    dn <- list()
    dn[[firstName]] <- labels
    if (!is.null(dimnames(source)))
        dn <- c(dn, dimnames(source))
    else
        dn <- c(dn, rep(list(NULL), length(.dims(source))))
    
    return (dn)
}

.replaceProperty <- function (shades, replacement, space, dim, name)
{
    UseMethod(".replaceProperty")
}

.replaceProperty.function <- function (shades, replacement, space, dim, name)
{
    function (...) .replaceProperty(shades(...), replacement, space, dim, name)
}

.replaceProperty.Scale <- function (shades, replacement, space, dim, name)
{
    ggplot2::ggproto(NULL, shades, palette=function(self,...) {
        colours <- ggplot2::ggproto_parent(shades, self)$palette(...)
        .replaceProperty(colours, replacement, space, dim, name)
    })
}

.replaceProperty.default <- function (shades, replacement, space, dim, name)
{
    shades <- warp(shades, space)
    
    if (is.null(replacement))
        return (unname(coords(shades)[,dim]))
    else
    {
        if (is.numeric(replacement) || is.logical(replacement))
        {
            arity <- length(replacement)
            valueLabels <- as.character(replacement)
            replacement <- rep(replacement, length(shades))
        }
        else
        {
            fun <- match.fun(replacement)
            arity <- length(fun(coords(shades)[1,dim]))
            replacement <- fun(coords(shades)[,dim])
            valueLabels <- NULL
        }
        
        missing <- is.na(replacement)
        indices <- rep(seq_along(shades), each=arity)
        coords <- coords(shades)[indices,,drop=FALSE]
        coords[!missing,dim] <- replacement[!missing]
        coords <- .clip(coords, space)
        alpha <- .alpha(shades, allowNull=FALSE)[indices]
        
        result <- drop(structure(shade(coords,space=space,alpha=alpha), dim=c(arity,.dims(shades))))
        dimnames(result) <- .makeDimnames(shades, result, valueLabels, name)
        return (result)
    }
}

#' Query or change colour properties
#' 
#' These functions obtain the value of a colour property, or modify it. They
#' will convert between colour spaces as required, but the RGB representation
#' will be appropriately updated in the result.
#' 
#' Brightness and lightness differ technically, in the sense that one is
#' absolute and the other is relative. Intuitively, a grey square on white
#' paper is brighter under bright sunlight than in a dark room, but its
#' lightness (relative to the white around it) is constant between conditions.
#' In these functions, brightness is "value" in HSV space and is between 0
#' and 1, while lightness is defined in Lab space and is between 0 and 100.
#' Luminance is different again, being more physically based and defined in XYZ
#' space. Saturation and chroma are also related. Hue is defined in HSV space,
#' with red at 0º (and 360º), which is the most familiar parameterisation.
#' 
#' @param shades One or more colours, in any suitable form (see
#'   \code{\link{shade}}), or a palette function or scale.
#' @param values New values for the property in question, with \code{NA} as a
#'   pass-through value that will leave the property as-is. If \code{NULL}, the
#'   current value(s) will be returned. May also be a function computing new
#'   values from old ones, such as \code{delta}, which adds its argument, or
#'   \code{scalefac}, which multiplies it.
#' @param ... Arguments to replacement functions \code{delta}, \code{scalefac}
#'   and \code{recycle}, which will be concatenated.
#' @return Current colour property values, or new colours of class
#'   \code{"shade"}. If \code{shades} is a function, the result will be a new
#'   function that wraps the old one and modifies its return value accordingly.
#' 
#' @note The colour property functions are vectorised over both of their
#'   arguments, such that the dimensions of the result will be
#'   \code{c(length(values),dim(shades))}. However, the \code{recycle} function
#'   can be used to suppress the usual dimensional expansion, and instead
#'   follow R's standard recycling rule.
#' 
#' @examples
#' saturation(c("papayawhip","lavenderblush","olivedrab"))
#' 
#' saturation("papayawhip", 0.7)
#' saturation("papayawhip", delta(0.2))
#' saturation("papayawhip", scalefac(1.5))
#' 
#' saturation(c("red","green"), c(0.4,0.6))
#' saturation(c("red","green"), recycle(0.4,0.6))
#' @author Jon Clayden <code@@clayden.org>
#' @rdname properties
#' @export
saturation <- function (shades, values = NULL)
{
    .replaceProperty(shades, values, "HSV", 2, "saturation")
}

#' @rdname properties
#' @export
brightness <- function (shades, values = NULL)
{
    .replaceProperty(shades, values, "HSV", 3, "brightness")
}

#' @rdname properties
#' @export
lightness <- function (shades, values = NULL)
{
    .replaceProperty(shades, values, "Lab", 1, "lightness")
}

#' @rdname properties
#' @export
luminance <- function (shades, values = NULL)
{
    .replaceProperty(shades, values, "XYZ", 2, "luminance")
}

#' @rdname properties
#' @export
chroma <- function (shades, values = NULL)
{
    .replaceProperty(shades, values, "LCh", 2, "chroma")
}

#' @rdname properties
#' @export
hue <- function (shades, values = NULL)
{
    .replaceProperty(shades, values, "HSV", 1, "hue")
}

#' @rdname properties
#' @export
opacity <- function (shades, values = NULL)
{
    # Handle functions and ggplot2 scales
    if (is.function(shades))
        return (function (...) opacity(shades(...), values))
    else if (inherits(shades, "Scale"))
    {
        result <- ggplot2::ggproto(NULL, shades, palette=function(self,...) {
            colours <- ggplot2::ggproto_parent(shades, self)$palette(...)
            opacity(colours, values)
        })
        return (result)
    }
    
    # From here on we're dealing with literal colours
    shades <- shade(shades)
    
    if (is.null(values))
        return (.alpha(shades, allowNull=FALSE))
    else
    {
        if (is.numeric(values) || is.logical(values))
        {
            arity <- length(values)
            valueLabels <- as.character(values)
            values <- rep(values, length(shades))
        }
        else
        {
            fun <- match.fun(values)
            arity <- length(fun(.alpha(shades,allowNull=FALSE)[1]))
            values <- fun(.alpha(shades, allowNull=FALSE))
            valueLabels <- NULL
        }

        indices <- rep(seq_along(shades), each=arity)
        coords <- coords(shades)[indices,,drop=FALSE]
        values[is.na(values)] <- .alpha(shades, allowNull=FALSE)[indices][is.na(values)]
        
        result <- drop(structure(shade(coords,space=space(shades),alpha=values), dim=c(arity,.dims(shades))))
        dimnames(result) <- .makeDimnames(shades, result, valueLabels, "opacity")
        return (result)
    }
}

#' @rdname properties
#' @export
delta <- function (...)
{
    values <- as.numeric(c(...))
    return (function(x) x+values)
}

#' @rdname properties
#' @export
scalefac <- function (...)
{
    values <- as.numeric(c(...))
    return (function(x) x*values)
}

#' @rdname properties
#' @export
recycle <- function (...)
{
    values <- as.numeric(c(...))
    return (function(x) rep(values,length.out=length(x)))
}

using("shades")

expect_equal(opacity("red"), 1)
expect_equal(opacity("#FF000080"), 128/255)
expect_equal(opacity(saturation("#FF000080",0.5)), 128/255)
expect_equal_shades(warp("#FF000080","HSV"), "#FF000080")

expect_equal_shades(opacity("red",c(0,0.5)), c("#FF000000","#FF000080"))
expect_equal_shades(opacity("red",delta(-0.5)), "#FF000080")
expect_null(attr(opacity("#FF000080",delta(0.5)),"alpha"))

shades <- shade(c("red","green","blue"))
matrix <- rep(shades, 2)
dim(matrix) <- c(3L,2L)

expect_null(attr(shades,"alpha"))
expect_equal(opacity(c(shades,"#00000000")), c(1,1,1,0))
shades[2] <- "#00FF0000"
expect_equal(attr(shades,"alpha"), c(1,0,1))
expect_equal(dim(opacity(matrix,0.5)), c(3L,2L))
expect_equal(dim(opacity(matrix,recycle(0.4,0.6))), c(3L,2L))

expect_true(shade("#FF8080") == shade("#FF8080FF"))
expect_false(shade("#FF8080") == shade("#FF8080DD"))
expect_true(grepl("Alpha values do not match", all.equal(shade("#FF8080"), "#FF8080DD")))

using("shades")

# Extract colour properties
expect_equal(saturation("red"), 1)
expect_equal(brightness("red"), 1)
expect_equal(saturation("grey40"), 0)
expect_equal(brightness(c("grey40","grey60")), c(0.4,0.6))
expect_equal(round(chroma(c("black","white","red"))), c(0,0,105))
expect_equal(round(lightness(c("black","white","red"))), c(0,100,53))

# Manipulate colour properties
expect_equal_shades(saturation("red",0.5), shade("#FF8080"))
expect_equal_shades(brightness("red",0.5), shade("#800000"))
expect_equal_shades(hue("red",delta(240)), shade("#0000FF"))
expect_equal_shades(hue("blue",delta(240)), shade("#00FF00"))
expect_equal_shades(lightness("red",scalefac(0.5)), shade("#A60000"))
expect_equal_shades(brightness("red",scalefac(0.4,0.6)), c("#660000","#990000"))
expect_equivalent(coords(brightness("grey40",c(0.2,0.6))), matrix(c(0,0,0,0,0.2,0.6),nrow=2))

# Manipulate colour matrices and check dimensions
shades <- shade(c("red","green","blue"))
matrix <- rep(shades, 2)
dim(matrix) <- c(3L,2L)

expect_null(dim(brightness(shades,0.5)))
expect_equal(length(brightness(shades,0.5)), 3L)
expect_equal(dim(brightness(shades,c(0.4,0.6))), c(2L,3L))
expect_equal_shades(brightness(shades,c(0.4,0.6)), c("#660000","#990000","#006600","#009900","#000066","#000099"))
expect_equal(dim(brightness(matrix,c(0.4,0.6))), c(2L,3L,2L))
expect_equal(dim(brightness(matrix,recycle(0.4,0.6))), c(3L,2L))

using("shades")

primaries <- shade(c("red", "green", "blue"))
expect_equal_shades(dichromat(primaries), c("#4D4222","#FFF600","#0027FF"))
expect_equal_shades(dichromat(primaries,"protanopic"), c("#4D4222","#FFF600","#0027FF"))
expect_equal_shades(dichromat(primaries,"deuteranopic"), c("#C5A700","#D0B335","#0068FE"))
expect_equal_shades(dichromat(primaries,"tritanopic"), c("#FF0050","#00FBFF","#372828"))

#' Simple colour swatches
#'
#' This function provides a simple visualisation of a colour series as a series
#' of boxes against the specified background colour. If the input has more than
#' one dimension then the boxes will be arranged in a grid (flattening further
#' dimensions after the second).
#' 
#' @param x One or more colours, in any suitable form (see [shade()]).
#' @param bg A background colour.
#' @param border The border colour to draw around each box.
#' @param cex The font size to use for labelling named shades.
#' @param font The font weight to use for labelling named shades, in the sense
#'   of [par()]. The default, 2, corresponds to boldface.
#' @param ... Additional arguments (currently unused).
#' 
#' @examples
#' swatch(c("red", "green", "blue"))
#' @seealso [par()]
#' @author Jon Clayden <code@@clayden.org>
#' @export
swatch <- function (x, bg = "white", border = "grey50", cex = 0.7, font = 2, ...)
{
    shades <- shade(x)
    grid <- .dims(shades, collapse=TRUE)
    if (length(grid) == 1L)
        grid <- c(grid, 1)
    
    width <- 0.9 / (max(grid) + 1)
    gap <- 1 / (max(grid) + 1)
    
    # Generate one centre value per location in each dimension, reverse the
    # second so that the plot "reads" top-to-bottom, and expand to one x and y
    # position per shade
    stops <- lapply(grid, function(i) gap * ((max(grid) - i) / 2 + seq_len(i)))
    stops[[2]] <- rev(stops[[2]])
    centres <- as.matrix(expand.grid(stops))
    
    dn <- dimnames(shades)
    if (is.null(dn))
        dn <- list(NULL, NULL)
    else if (length(dn) == 1L)
        dn <- c(dn, list(NULL))
    named <- !vapply(dn, is.null, logical(1L))
    dimLabels <- .names(dn, allowNull=FALSE)
    
    oldPars <- par(mai=c(0,0,0,0), bg=bg)
    on.exit(par(oldPars))
    
    devSize <- dev.size()
    devRatio <- devSize[2] / devSize[1]
    
    xlim <- c(-0.1 - ifelse(named[2],1.5*width,0), 1.1)
    ylim <- 0.5 + c(-1,1) * devRatio * c(0.6 + ifelse(named[1],1.5*width,0), 0.6)
    
    plot(NA, NA, xlim=xlim, ylim=ylim, xlab="", ylab="", xaxt="n", yaxt="n", bty="n", asp=1)
    rect(centres[,1]-width/2, centres[,2]-width/2, centres[,1]+width/2, centres[,2]+width/2, col=shades, border=border, lwd=2)
    
    if (named[1])
    {
        text(stops[[1]], min(centres[,2])-width, dn[[1]], col="grey40")
        text(mean(centres[,1]), min(centres[,2])-1.5*width, dimLabels[1], font=2, xpd=TRUE)
    }
    if (named[2])
    {
        text(min(centres[,1])-width, stops[[2]], dn[[2]], col="grey40", srt=90)
        text(min(centres[,1])-1.5*width, mean(centres[,2]), dimLabels[2], font=2, srt=90, xpd=TRUE)
    }
    
    if (.hasNames(shades))
    {
        blackContrast <- contrast(shades, "black")
        whiteContrast <- contrast(shades, "white")
        labelCol <- ifelse(blackContrast > whiteContrast, "black", "white")
        text(centres[,1], centres[,2], names(shades), col=labelCol, cex=cex, font=font)
    }
}

#' Expectation for comparing shades
#' 
#' This function provides an expectation for use with the `tinytest` package,
#' which compares two colour vectors via `all.equal.shade(hexonly=TRUE)`. It
#' evaluates `TRUE` if the vectors have the same length and resolve to the same
#' hex colour strings.
#' 
#' @param current The colour vector to test.
#' @param target The target values to compare against, in any form valid for
#'   casting to [shade()].
#' @param ... Further arguments to [all.equal()].
#' @param info An optional information string shown in case of failure.
#' @return A `"tinytest"` object.
#' 
#' @rdname tinytest
#' @export
expect_equal_shades <- function (current, target, ..., info = NA_character_)
{
    result <- all.equal(shade(target), shade(current), hexonly=TRUE, ...)
    diff <- (if (isTRUE(result)) NA_character_ else result)
    return (tinytest::tinytest(isTRUE(result), call=sys.call(sys.parent(1)), diff=diff, short="data", info=info))
}

# The Brettel algorithm calls for "the brightest possible metamer of an equal-
# energy stimulus". The equal-energy stimulus, CIE illuminant E, has equal
# values for each coordinate in XYZ space. 0.8388 is the largest equal-energy
# value (to 4 d.p.) within the sRGB gamut, according to my experimentation.
.equalEnergy <- matrix(0.8388, nrow=1, ncol=3)

# XYZ space coordinates for the specified wavelengths, using the standard (CIE
# 1931) observer. This is a small subset of the values available from
# http://files.cie.co.at/204.xls.
.standardObserver <- list("475"=matrix(c(0.1421,0.1126,1.0419), nrow=1),
                          "485"=matrix(c(0.05795,0.1693,0.6162), nrow=1),
                          "575"=matrix(c(0.8425,0.9154,0.0018), nrow=1),
                          "660"=matrix(c(0.1649,0.061,0), nrow=1))

# Derived coefficients from the values above, calculated and cached when the
# package namespace is loaded
.cache <- new.env()

#' Simulate colour appearance for dichromats
#' 
#' This functions manipulates colours to simulate the effects of different
#' kinds of colour blindness, and specifically dichromacy, in which only two of
#' the usual three types of photoreceptors are present. There are three types,
#' corresponding to the loss of red, green or blue photoreceptors.
#' 
#' @param shades One or more colours, in any suitable form (see
#'   \code{\link{shade}}).
#' @param type The type of colour vision deficiency to simulate: protanopia
#'   (red blindness), deuteranopia (green blindness) or tritanopia (blue
#'   blindness). The latter is the rarest in the population. \code{"none"} is
#'   also available, as a pass-through option. Abbreviations, such as the first
#'   letter, may be used, and multiple values are acceptable.
#' @return New colours of class \code{"shade"} in LMS space, representing
#'   projections of the original shades onto a submanifold appropriate to the
#'   type of dichromacy being simulated.
#' 
#' @examples
#' dichromat(c("red", "green", "blue"))
#' @references
#' Brettel, H., Viénot, F. and Mollon, J.D. (1997). Computerized simulation of
#' color appearance for dichromats. Journal of the Optical Society of America A
#' 14(10):2647-2655.
#' @author Jon Clayden <code@@clayden.org>
#' @export
dichromat <- function (shades, type = c("protanopic","deuteranopic","tritanopic","none"))
{
    if (missing(type))
        type <- "protanopic"
    else
        type <- match.arg(type, several.ok=TRUE)
    
    shades <- warp(shades, "LMS")
    Q <- coords(shades)
    coords <- do.call(rbind, lapply(type, function(t) {
        Qprime <- Q
        if (t != "none")
        {
            lambda <- switch(t, protanopic=ifelse(Q[,3]/Q[,2] < .cache$Er[1], 575, 475),
                                deuteranopic=ifelse(Q[,3]/Q[,1] < .cache$Er[2], 575, 475),
                                tritanopic=ifelse(Q[,2]/Q[,1] < .cache$Er[3], 660, 485))
            lambda <- as.character(lambda)
            
            if (t == "protanopic")
                Qprime[,1] <- -(.cache$b[lambda]*Q[,2] + .cache$c[lambda]*Q[,3]) / .cache$a[lambda]
            else if (t == "deuteranopic")
                Qprime[,2] <- -(.cache$a[lambda]*Q[,1] + .cache$c[lambda]*Q[,3]) / .cache$b[lambda]
            else
                Qprime[,3] <- -(.cache$a[lambda]*Q[,1] + .cache$b[lambda]*Q[,2]) / .cache$c[lambda]
        }
        return (Qprime)
    }))
    
    indices <- rep(seq_along(shades),each=length(type)) + length(shades) * (seq_along(type)-1)
    coords <- coords[indices,,drop=FALSE]
    
    return (drop(structure(shade(coords,space="LMS"), dim=c(length(type),.dims(shades)))))
}

#' Colour distance and contrast
#' 
#' The `distance()` function calculates a distance measure that aims to
#' quantify perceptual difference between a vector of colours and a reference
#' colour. The measure in question is the CIE Delta E (2000), which is
#' calculated based on colour coordinates in Lab space. Colour contrast relates
#' more specifically to legibility, in the context where text in one colour is
#' overlaid on a background in another. The W3C definition used here ranges
#' between 1 (identical colours, no contrast) and 21 (black on white, maximum
#' contrast).
#' 
#' @param shades One or more colours, in any suitable form (see [shade()]).
#' @param reference A single reference colour.
#' @return A numeric vector of distances or contrasts.
#' 
#' @examples
#' distance(c("red","green","blue"), "red")
#' contrast(c("red","green","blue"), "red")
#' @references
#' See \url{http://www.brucelindbloom.com/index.html?Eqn_DeltaE_CIE2000.html}
#' for CIE Delta E. The contrast measure is defined as part of the W3C Web
#' Content Accessibility Guidelines, version 2.0, detailed at
#' \url{https://www.w3.org/TR/WCAG20/}.
#' @author Jon Clayden <code@@clayden.org>
#' @export
distance <- function (shades, reference)
{
    deg2rad <- function(x) x / 180 * pi
    rad2deg <- function(x) x / pi * 180
    
    shadeCoords <- coords(warp(shades, "Lab"))
    L2 <- shadeCoords[,1]
    a2 <- shadeCoords[,2]
    b2 <- shadeCoords[,3]
    
    if (length(reference) != 1L)
        stop("Reference should be a single shade")
    
    refCoords <- coords(warp(reference, "Lab"))
    L1 <- refCoords[,1]
    a1 <- refCoords[,2]
    b1 <- refCoords[,3]
    
    LBarPrime <- (L1 + L2) / 2
    C1 <- sqrt(a1^2 + b1^2)
    C2 <- sqrt(a2^2 + b2^2)
    CBar <- (C1 + C2) / 2
    G <- (1 - sqrt(CBar^7/(CBar^7+25^7))) / 2
    a1Prime <- a1 * (1+G)
    a2Prime <- a2 * (1+G)
    C1Prime <- sqrt(a1Prime^2 + b1^2)
    C2Prime <- sqrt(a2Prime^2 + b2^2)
    CBarPrime <- (C1Prime + C2Prime) / 2
    h1Prime <- atan2(b1, a1Prime) %% (2*pi)
    h2Prime <- atan2(b2, a2Prime) %% (2*pi)
    largeDiff <- abs(h1Prime-h2Prime) > pi
    HBarPrime <- (h1Prime + h2Prime + ifelse(largeDiff, 2*pi, 0)) / 2
    bigT <- 1 - 0.17*cos(HBarPrime-deg2rad(30)) + 0.24*cos(2*HBarPrime) + 0.32*cos(3*HBarPrime+deg2rad(6)) - 0.20*cos(4*HBarPrime-deg2rad(63))
    deltahPrime <- h2Prime - h1Prime + ifelse(largeDiff, ifelse(h2Prime>h1Prime,-2*pi,2*pi), 0)
    deltaLPrime <- L2 - L1
    deltaCPrime <- C2Prime - C1Prime
    deltaHPrime <- 2 * sqrt(C1Prime*C2Prime) * sin(deltahPrime/2)
    SL <- 1 + 0.015 * (LBarPrime-50)^2 / sqrt(20 + (LBarPrime-50)^2)
    SC <- 1 + 0.045 * CBarPrime
    SH <- 1 + 0.015 * CBarPrime * bigT
    deltatheta <- deg2rad(30 * exp(-((rad2deg(HBarPrime)-275) / 25)^2))
    RC <- 2 * sqrt(CBarPrime^7 / (CBarPrime^7 + 25^7))
    RT <- -RC * sin(2*deltatheta)
    
    unname(sqrt((deltaLPrime/SL)^2 + (deltaCPrime/SC)^2 + (deltaHPrime/SH)^2 + RT*(deltaCPrime/SC)*(deltaHPrime/SH)))
}

#' @rdname distance
#' @export
contrast <- function (shades, reference)
{
    Y <- luminance(shades)
    Yref <- luminance(reference)
    (pmax(Y,Yref) + 0.05) / (pmin(Y,Yref) + 0.05)
}

expect_equal(round(distance(c("red","green","blue"),"red")), c(0,87,53))
expect_error(distance("red", c("red","green","blue")), "Reference should be a single shade")
expect_equal(contrast("black","white"), 21)

# Contributing

This document is the official contribution guide contributors must follow. It will be **greatly appreciated** if you
read it first before contributing. It will also prevent you from losing your time if you open an issue / make a PR that
doesn’t comply to this document.

<!-- vim-markdown-toc GFM -->

* [Disclaimer and why this document](#disclaimer-and-why-this-document)
* [How to make a change](#how-to-make-a-change)
  * [Process](#process)
* [Conventions](#conventions)
  * [Coding](#coding)
  * [Git](#git)
    * [Git message](#git-message)
    * [Commit atomicity](#commit-atomicity)
    * [Hygiene](#hygiene)
* [Release process](#release-process)
  * [Overall process](#overall-process)
  * [Changelogs update](#changelogs-update)
  * [Git tag](#git-tag)
* [Support and donation](#support-and-donation)

<!-- vim-markdown-toc -->

# Disclaimer and why this document

People contributing is awesome. The more people contribute to Free & Open-Source software, the better the
world is to me. However, the more people contribute, the more work we have to do on our spare-time. Good
contributions are highly appreciated, especially if they thoroughly follow the conventions and guidelines of
each and every repository. However, bad contributions — that don’t follow this document, for instance — are
going to require me more work than was involved into making the actual change. It’s even worse when the contribution
actually solves a bug or add a new feature.

So please read this document; it’s not hard and the few rules here are easy to respect. You might already do
everything in this list anyway, but reading it won’t hurt you. For more junior / less-experienced developers, it’s
very likely you will learn a bit of process that is considered good practice, especially when working with VCS like
Git.

> Thank you!

# How to make a change

## Process

The typical process is to base your work on the `master` branch. The `master` branch must always contain a stable
version of the project. It is possible to make changes by basing your work on other branches but the source
of truth is `master`. If you want to synchronize with other people on other branches, feel free to.

The process is:

1. (optional) Open an issue and discuss what you want to do. This is optional but highly recommended. If you
  don’t open an issue first and work on something that is not in the scope of the project, or already being
  made by someone else, you’ll be working for nothing. Also, keep in mind that if your change doesn’t refer to an
  existing issue, I will be wondering what is the context of your change. So prepare to be asked about the motivation
  and need behind your changes — it’s greatly appreciated if the commit messages, code and PR’s content already
  contains this information so that people don’t have to ask.
2. Fork the project.
3. Create a branch starting from `master` – or the branch you need to work on. Even though this is not really enforced,
  you’re advised to name your branch according to the _Git Flow_ naming convention:
  - `fix/your-bug-here`: if you’re fixing a bug, name your branch.
  - `feature/new-feature-here`: if you’re adding some work.
  - Free for anything else.
  - The special `release/*` branch is used to either back-port changes from newer versions to previous
    versions, or to release new versions by updating `Cargo.toml` files, changelogs, etc. Normally, contributors should
    never have to worry about this kind of brach as their creations is often triggered when wanting to make a release.
4. Make some commits!
5. Once you’re ready, open a Pull Request (PR) to merge your work on the target branch. For instance, open a PR for
  `master <- feature/something-new`.
6. (optional) Ask someone to review your code in the UI. Normally, I’m pretty reactive to notifications but it never
  hurts to ask for a review.
7. Discussion and peer-review.
8. Once the CI is all green, someone (likely me [@phaazon]) will merge your code and close your PR.
9. Feel free to delete your branch.

# Conventions

## Coding

Coding conventions are enforced by `rustfmt`. You are expected to provide code that is formatted by `rustfmt`
with the top-level `rustfmt.toml` file.

Coding convention is enforced in the Continuous Integration pipeline. If you want your work to be
mergeable, format your code.

> Note: please do not format your code in a separate, standalone commit. This way of doing is
> considered as a bad practice as the commit will not contain _anything_ useful (but code
> reformatted). Please format all your commits. You can use various tools in your editor to do so,
> such as [rust-analyzer](https://github.com/rust-analyzer/rust-analyzer).

## Git

### Git message

Please format your git messages like so:

> [project(s)] Starting with an uppercase letter, ending with a dot. #343
>
> The #343 after the dot is appreciated to link to issues. Feel free to add, like this message, more context
> and/or precision to your git message. You don’t have to put it in the first line of the commit message,
> but if you are fixing a bug or implementing a feature thas has an issue linked, please reference it, so
> that it is easier to generate changelogs when reading the git log.

The `[project(s)]` header is mandatory if your commit has changes in any of the crates handled by this repository.
If the repository you contribute to has a single project in it, you can omit the `[project(s)]` part. If you make
a change that is cross-crate, feel free to separate them with commas, like `[crate_a, crate_b]`. If
you make a change that touches all the crates, you can use `[all]`. If you change something that is not related
to a crate, like the front README, CONTRIBUTING file, CI setup, top-level `Cargo.toml`, etc., then you can omit
this header.

**I’m very strict on git messages as I use them to write `CHANGELOG.md` files. Don’t be surprised if I ask you
to edit a commit message. :)**

### Commit atomicity

Your commits should be as atomic as possible. That means that if you make a change that touches two different
concepts / has different scopes, most of the time, you want two commits – for instance one commit for the backend crate
and one commit for the interface crate. There are exceptions, so this is not an absolute rule, but take some time
thinking about whether you should split your commits or not. Commits which add a feature / fix a bug _and_ add tests at
the same time are fine.

However, here’s a non-comprehensive list of commits considered bad and that will be refused:

- **Formatting, refactoring, cleaning, linting code in a PR that is not strictly about formatting**. If you open a PR to
  fix a bug, implement a  feature, change configuration, add metadata to the CI, etc. — pretty much anything — but you
  also format some old code that has nothing to do with your PR, apply a linter’s suggestions (such as `clippy`), remove
  old code, etc., then I will refuse your commit(s) and ask you to edit your PR.
- **Too atomic commits**. If two commits are logically connected to one another and are small, it’s likely that you want
  to merge them as a single commit — unless they work on too different parts of your code. This is a bit subjective
  topic, so I won’t be _too picky_ about it, but if I judge that you should split a commit into two or fixup two commits,
  please don’t take it too personal. :)

If you don’t know how to write your commits in an atomic maneer, think about how one would revert your commits if
something bad happens with your changes — like a big breaking change we need to roll back from very quickly. If your
commits are not atomic enough, rolling them back will also roll back code that has nothing to do with your changes.

### Hygiene

When working on a fix or a feature, it’s very likely that you will periodically need to update your branch
with the `master` branch. **Do not use merge commits**, as your contributions will be refused if you have
merge commits in them. The only case where merge commits are accepted is when you work with someone else
and are required to merge another branch into your feature branch (and even then, it is even advised to
simply rebase). If you want to synchronize your branch with `master`, please use:

```
git switch <your_branch>
git fetch origin --prune
git rebase origin/master
```

# Release process

## Overall process

Releases occur at arbitrary rates. If something is considered urgent, it is most of the time released immediately
after being merged and tested. Sometimes, several issues are being fixed at the same time (spanning on a few
days at max). Those will be gathered inside a single update.

Feature requests might be delayed a bit to be packed together as well but eventually get released, even if
they’re small. Getting your PR merged means it will be released _soon_, but depending on the urgency of your changes,
it might take a few minutes to a few days.

## Changelogs update

`CHANGELOG.md` files must be updated **before any release**. Especially, they must contain:

- The version of the release.
- The date of the release.
- How to migrate from a minor to the next major.
- Everything that a release has introduced, such as major, minor and patch changes.

Because I don’t ask people to maintain changelogs, I have a high esteem of people knowing how to use Git and create
correct commits. Be advised that I will refuse any commit that prevents me from writing the changelog correctly.

## Git tag

Once a new release occurs, a Git tag is created. Git tags are formatted regarding the project they refer to, if several
projects are present in the repository. If only one project is present, tags will refer to this project by the same
naming scheme anyway:

> <project-name>-X.Y.Z

Where `X` is the _major version_, `Y` is the _minor version_ and `Z` is the _patch version_. For instance
`project-0.37.1` is a valid Git tag, so is `project-derive-0.5.3`.

A special kind of tag is also possible:

> <project-name>-X.Y.Z-rc.W

Where `W` is a number starting from `1` and incrementing. This format is for _release candidates_ and occurs
when a new version (most of the time a major one) is to be released but more feedback is required.

Crates are pushed to [crates.io](https://crates.io) and tagged on Git manually (no CD involved).

# Support and donation

This project is a _free  and  open-source_ project. It has no financial motivation nor support. I
([@phaazon]) would like to make it very clear that:

- Sponsorship is not available. You cannot pay me to make me do things for you. That includes issues reports,
  features requests and such.
- If you still want to donate because you like the project and think I should be rewarded, you are free to
  give whatever you want.
- However, keep in mind that donating doesn’t unlock any privilege people who don’t donate wouldn’t already
  have. This is very important as it would bias priorities. Donations must remain anonymous.
- For this reason, no _sponsor badge_ will be shown, as it would distinguish people who donate from those
  who don’t. This is a _free and open-source_ project, everybody is welcome to contribute, with or without
  money.

[@phaazon]: https://github.com/phaazon

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/properties.R
\name{saturation}
\alias{saturation}
\alias{brightness}
\alias{lightness}
\alias{luminance}
\alias{chroma}
\alias{hue}
\alias{opacity}
\alias{delta}
\alias{scalefac}
\alias{recycle}
\title{Query or change colour properties}
\usage{
saturation(shades, values = NULL)

brightness(shades, values = NULL)

lightness(shades, values = NULL)

luminance(shades, values = NULL)

chroma(shades, values = NULL)

hue(shades, values = NULL)

opacity(shades, values = NULL)

delta(...)

scalefac(...)

recycle(...)
}
\arguments{
\item{shades}{One or more colours, in any suitable form (see
\code{\link{shade}}), or a palette function or scale.}

\item{values}{New values for the property in question, with \code{NA} as a
pass-through value that will leave the property as-is. If \code{NULL}, the
current value(s) will be returned. May also be a function computing new
values from old ones, such as \code{delta}, which adds its argument, or
\code{scalefac}, which multiplies it.}

\item{...}{Arguments to replacement functions \code{delta}, \code{scalefac}
and \code{recycle}, which will be concatenated.}
}
\value{
Current colour property values, or new colours of class
\code{"shade"}. If \code{shades} is a function, the result will be a new
function that wraps the old one and modifies its return value accordingly.
}
\description{
These functions obtain the value of a colour property, or modify it. They
will convert between colour spaces as required, but the RGB representation
will be appropriately updated in the result.
}
\details{
Brightness and lightness differ technically, in the sense that one is
absolute and the other is relative. Intuitively, a grey square on white
paper is brighter under bright sunlight than in a dark room, but its
lightness (relative to the white around it) is constant between conditions.
In these functions, brightness is "value" in HSV space and is between 0
and 1, while lightness is defined in Lab space and is between 0 and 100.
Luminance is different again, being more physically based and defined in XYZ
space. Saturation and chroma are also related. Hue is defined in HSV space,
with red at 0º (and 360º), which is the most familiar parameterisation.
}
\note{
The colour property functions are vectorised over both of their
arguments, such that the dimensions of the result will be
\code{c(length(values),dim(shades))}. However, the \code{recycle} function
can be used to suppress the usual dimensional expansion, and instead
follow R's standard recycling rule.
}
\examples{
saturation(c("papayawhip","lavenderblush","olivedrab"))

saturation("papayawhip", 0.7)
saturation("papayawhip", delta(0.2))
saturation("papayawhip", scalefac(1.5))

saturation(c("red","green"), c(0.4,0.6))
saturation(c("red","green"), recycle(0.4,0.6))
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shade.R
\name{warp}
\alias{warp}
\title{Shift colours between spaces}
\usage{
warp(x, space)
}
\arguments{
\item{x}{An R object which can be coerced to class \code{"shade"}.}

\item{space}{A string naming the new space.}
}
\value{
A new object of class \code{"shade"}.
}
\description{
This function shifts the current colour space of its arguments to the
specified space, returning a new object of class \code{"shade"}.
}
\details{
Valid names for spaces are currently those supported by the
\code{\link{convertColor}} function, namely "sRGB", "Apple RGB", "CIE
RGB", "XYZ", "Lab" and "Luv"; plus "RGB" (which is treated as an
alias for "sRGB"), "HSV", "LCh", "LMS" and "Oklab". Case is not significant.
}
\note{
LMS space, used for chromatic adaptation and simulating colour
blindness, is not uniquely defined. Here we use the (linearised) Bradford
transform, obtained by Lam (1985) and used widely in ICC colour profiles
and elsewhere, to transform to and from CIE XYZ space.

R uses the D65 standard illuminant as the reference white for the "Lab"
and "Luv" spaces.
}
\examples{
warp("red", "HSV")
}
\references{
The Oklab colour space is explained at \url{https://bottosson.github.io/posts/oklab/}.

Lam, K.M. (1985). Metamerism and colour constancy. PhD thesis, University of
Bradford.
}
\seealso{
\code{\link{convertColor}}
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/swatch.R
\name{swatch}
\alias{swatch}
\title{Simple colour swatches}
\usage{
swatch(x, bg = "white", border = "grey50", cex = 0.7, font = 2, ...)
}
\arguments{
\item{x}{One or more colours, in any suitable form (see \code{\link[=shade]{shade()}}).}

\item{bg}{A background colour.}

\item{border}{The border colour to draw around each box.}

\item{cex}{The font size to use for labelling named shades.}

\item{font}{The font weight to use for labelling named shades, in the sense
of \code{\link[=par]{par()}}. The default, 2, corresponds to boldface.}

\item{...}{Additional arguments (currently unused).}
}
\description{
This function provides a simple visualisation of a colour series as a series
of boxes against the specified background colour. If the input has more than
one dimension then the boxes will be arranged in a grid (flattening further
dimensions after the second).
}
\examples{
swatch(c("red", "green", "blue"))
}
\seealso{
\code{\link[=par]{par()}}
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shade.R
\name{shade}
\alias{shade}
\alias{shade.shade}
\alias{shade.color}
\alias{shade.matrix}
\alias{shade.character}
\alias{shade.default}
\alias{shades}
\alias{print.shade}
\alias{[.shade}
\alias{[<-.shade}
\alias{c.shade}
\alias{rep.shade}
\alias{rev.shade}
\alias{==.shade}
\alias{!=.shade}
\alias{all.equal.shade}
\title{The shade class}
\usage{
shade(x, ...)

\method{shade}{shade}(x, ...)

\method{shade}{color}(x, ...)

\method{shade}{matrix}(x, space = "sRGB", alpha = NULL, ...)

\method{shade}{character}(x, ...)

\method{shade}{default}(x, ...)

shades(...)

\method{print}{shade}(x, ...)

\method{[}{shade}(x, i)

\method{[}{shade}(x, i) <- value

\method{c}{shade}(...)

\method{rep}{shade}(x, ...)

\method{rev}{shade}(x)

\method{==}{shade}(x, y)

\method{!=}{shade}(x, y)

\method{all.equal}{shade}(target, current, hexonly = FALSE, ...)
}
\arguments{
\item{x, y}{R objects, or \code{"shade"} objects for methods.}

\item{...}{For \code{shade()}, additional parameters to methods. For \code{shades()}
and \code{c()}, any number of colours in any acceptable form.}

\item{space}{For a matrix, the space in which coordinates are being
provided.}

\item{alpha}{For a matrix, an associated vector of opacity values between 0
and 1, if required.}

\item{i}{An index vector.}

\item{value}{A vector of replacement colours.}

\item{target, current}{Shade vectors to compare.}

\item{hexonly}{If \code{TRUE}, compare only on the basis of the hex strings.
Otherwise test for equal coordinates.}
}
\value{
A character vector of class \code{"shade"}, with additional attributes
as follows.
\item{space}{A string naming a colour space.}
\item{coords}{A matrix giving colour coordinates in the relevant space,
one colour per row.}
\item{alpha}{A numeric vector of alpha (opacity) values between 0 and 1,
one value per element of the vector. If not present then full opacity
is assumed for every element.}
\item{names}{A character vector of element names, in the usual R
convention. If not present then the object is unnamed.}
}
\description{
Objects of class \code{"shade"} are simply standard R character vectors
representing one or more 8-bit (s)RGB colours in CSS-like hex format, but
with extra attributes giving the current colour space and coordinates.
Opacity values are also supported.
}
\details{
The \code{shade()} function creates a colour vector from a single main argument
plus method-specific arguments where appropriate. The \code{shades()} function
provides an alternative interface that converts each of its arguments to
shades, with direct inline naming, and then concatenates them.

Comparison between \code{"shade"} objects \code{x} and \code{y} is achieved by converting
\code{y} (the second argument) into the colour space of \code{x} and then comparing
coordinates, after any clipping.
}
\note{
When concatenating, shades that are all from the same space will
remain in that space, but shades from different spaces will be warped to
"XYZ" space.
}
\examples{
s <- shade(c("red", "green", "blue"))
s[1]
s[1] <- "pink"
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixtures.R
\name{complement}
\alias{complement}
\title{Complementary colours}
\usage{
complement(shades, space = NULL)
}
\arguments{
\item{shades}{One or more colours, in any suitable form (see
\code{\link{shade}}), or a palette function or scale.}

\item{space}{The space in which to take the complement.}
}
\value{
New colours of class \code{"shade"}, or a new palette function.
}
\description{
This function returns the complement of its argument, the "opposite" colours
in the specified space.
}
\examples{
complement("cyan")
complement("cyan", space="HSV")
complement("cyan", space="Lab")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adaptation.R
\name{dichromat}
\alias{dichromat}
\title{Simulate colour appearance for dichromats}
\usage{
dichromat(shades, type = c("protanopic", "deuteranopic", "tritanopic", "none"))
}
\arguments{
\item{shades}{One or more colours, in any suitable form (see
\code{\link{shade}}).}

\item{type}{The type of colour vision deficiency to simulate: protanopia
(red blindness), deuteranopia (green blindness) or tritanopia (blue
blindness). The latter is the rarest in the population. \code{"none"} is
also available, as a pass-through option. Abbreviations, such as the first
letter, may be used, and multiple values are acceptable.}
}
\value{
New colours of class \code{"shade"} in LMS space, representing
projections of the original shades onto a submanifold appropriate to the
type of dichromacy being simulated.
}
\description{
This functions manipulates colours to simulate the effects of different
kinds of colour blindness, and specifically dichromacy, in which only two of
the usual three types of photoreceptors are present. There are three types,
corresponding to the loss of red, green or blue photoreceptors.
}
\examples{
dichromat(c("red", "green", "blue"))
}
\references{
Brettel, H., Viénot, F. and Mollon, J.D. (1997). Computerized simulation of
color appearance for dichromats. Journal of the Optical Society of America A
14(10):2647-2655.
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scales.R
\name{gradient}
\alias{gradient}
\title{Colour gradients}
\usage{
gradient(shades, steps = NULL, space = NULL)
}
\arguments{
\item{shades}{Two or more colours, in any suitable form (see
\code{\link{shade}}), or a named colour map such as \code{"viridis"}.}

\item{steps}{An integer giving the number of shades required in the palette.
If \code{NULL}, a function will instead be returned that takes this
argument.}

\item{space}{The colour space to traverse. Defaults to the current space of
\code{shades}, or \code{"Lab"} for the \code{matplotlib} colour maps, or
\code{"sRGB"} otherwise.}
}
\value{
A character vector of class \code{"shade"} containing the gradient
elements in the specified space, or a palette function.
}
\description{
This function returns a set of colours interpolating between the specified
key colours, equally separated in the specified space.
}
\details{
The key colours may be specified explicitly, or else a built-in colour map
may be used. The maps available are currently those developed for Python's
\code{matplotlib} 2.0, namely \code{"magma"}, \code{"inferno"},
\code{"plasma"} and \code{"viridis"}, and certain ColorBrewer palettes,
namely \code{"Blues"}, \code{"Reds"}, \code{"YlOrRd"} (yellow-orange-red)
and \code{"RdBu"} (red-grey-blue, a balanced diverging scale).
}
\examples{
gradient(c("red","blue"), 5)
gradient(c("red","blue"), 5, space="Lab")
gradient("viridis", 5)
}
\references{
\url{https://bids.github.io/colormap/} for the \code{matplotlib} colour maps;
\url{https://colorbrewer2.org} for the ColorBrewer ones.
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinytest.R
\name{expect_equal_shades}
\alias{expect_equal_shades}
\title{Expectation for comparing shades}
\usage{
expect_equal_shades(current, target, ..., info = NA_character_)
}
\arguments{
\item{current}{The colour vector to test.}

\item{target}{The target values to compare against, in any form valid for
casting to \code{\link[=shade]{shade()}}.}

\item{...}{Further arguments to \code{\link[=all.equal]{all.equal()}}.}

\item{info}{An optional information string shown in case of failure.}
}
\value{
A \code{"tinytest"} object.
}
\description{
This function provides an expectation for use with the \code{tinytest} package,
which compares two colour vectors via \code{all.equal.shade(hexonly=TRUE)}. It
evaluates \code{TRUE} if the vectors have the same length and resolve to the same
hex colour strings.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{distance}
\alias{distance}
\alias{contrast}
\title{Colour distance and contrast}
\usage{
distance(shades, reference)

contrast(shades, reference)
}
\arguments{
\item{shades}{One or more colours, in any suitable form (see \code{\link[=shade]{shade()}}).}

\item{reference}{A single reference colour.}
}
\value{
A numeric vector of distances or contrasts.
}
\description{
The \code{distance()} function calculates a distance measure that aims to
quantify perceptual difference between a vector of colours and a reference
colour. The measure in question is the CIE Delta E (2000), which is
calculated based on colour coordinates in Lab space. Colour contrast relates
more specifically to legibility, in the context where text in one colour is
overlaid on a background in another. The W3C definition used here ranges
between 1 (identical colours, no contrast) and 21 (black on white, maximum
contrast).
}
\examples{
distance(c("red","green","blue"), "red")
contrast(c("red","green","blue"), "red")
}
\references{
See \url{http://www.brucelindbloom.com/index.html?Eqn_DeltaE_CIE2000.html}
for CIE Delta E. The contrast measure is defined as part of the W3C Web
Content Accessibility Guidelines, version 2.0, detailed at
\url{https://www.w3.org/TR/WCAG20/}.
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixtures.R
\name{addmix}
\alias{addmix}
\alias{submix}
\alias{\%.)\%}
\alias{\%_/\%}
\title{Colour mixtures}
\usage{
addmix(base, mixer, amount = 1, space = NULL)

submix(base, mixer, amount = 1, space = NULL)

X \%.)\% Y

X \%_/\% Y
}
\arguments{
\item{base, X}{A vector of base colours, or a palette function or scale.}

\item{mixer, Y}{A vector of colours to mix in.}

\item{amount}{The amount of each colour to mix in, relative to the amount
of the base. This will be recycled to the length of \code{mixer}.}

\item{space}{A string giving the space in which to perform the mixing, or
\code{NULL}. In the latter case, the space of \code{base} will be used.}
}
\value{
New colours of class \code{"shade"}, or a new palette function.
}
\description{
These functions allow colours to be mixed in any colour space, either
additively (like light) or subtractively (like paint). The infix form
\code{\%.)\%} is an alternative for \code{addmix}, and \code{\%_/\%} for
\code{submix}, with the mixing amount being fixed to 1 in these cases.
}
\examples{
addmix(c("red","green","blue"), "red")
submix(c("cyan","magenta","yellow"), "cyan")
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shade.R
\name{space}
\alias{space}
\title{Retrieve the space of a colour vector}
\usage{
space(x, ...)
}
\arguments{
\item{x}{An R object.}

\item{...}{Additional arguments to methods.}
}
\value{
A string naming a colour space.
}
\description{
This function retrieves the colour space in which its argument is currently
defined.
}
\examples{
space("red")
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shade.R
\name{coords}
\alias{coords}
\title{Retrieve the coordinates of a colour vector}
\usage{
coords(x, ...)
}
\arguments{
\item{x}{An R object.}

\item{...}{Additional arguments to methods.}
}
\value{
A matrix giving colour coordinates in the relevant space, one colour
per row. Columns are typically named.
}
\description{
This function retrieves the coordinates of a colour vector's elements,
within whatever space it is currently defined.
}
\examples{
coords("red")
}
\author{
Jon Clayden \href{mailto:code@clayden.org}{code@clayden.org}
}
